

/*#define DEBUG_TCP*/

#include <nostub.h>
#include <alloc.h>
#include <string.h>
#include <link.h>
#include <kbd.h>
#include <graph.h>
#include <peekpoke.h>
#include <system.h>
#include <unknown.h>
#include <dialogs.h>
#include <menus.h>
#include <ctype.h>
#include <vat.h>
int _ti89;

#define True		1
#define False		0
#define Void void
#define Longword	unsigned long	/* 32 bits */
#define	Word		unsigned short	/* 16 bits */
#define	Byte		unsigned char	/* 8 bits */


#define	ADDR(a,b,c,d)	(((long)a<<24L)|((long)b<<16L)|((long)c<<8L)|((long)d))
#define MY_ADDR         ADDR( 192, 168, 0, 64 )
#define HOST_ADDR       ADDR( 192, 168, 0, 1 )
#define IP_Address	Longword

#define P(x) x
typedef void ( *Procref ) P((void *s, Byte *dp, int len ));
typedef void ( *Procrefv ) P(( void ));


struct	Ethernet_Address {
	Word		w[ 3 ];
};
/* The Ethernet header */

struct eth_Header {
	struct	Ethernet_Address destination;	/* 48 bits = 6 bytes */
	struct	Ethernet_Address source;	/* 48 bits = 6 bytes */
	Word		type;		/* 16 bits = 2 bytes */
};					/* total 14 bytes */


/* The Internet Header: */
struct in_Header {
	Word		vht;		/* version, hdrlen, tos */
					/* 2 bytes */
				/* version = first nybble = 4. */
				/* hdrlen = low nybble of 1st byte =
					length of header in dwords = 5 */
	Word		length;		/* 2 bytes */
	Word		identification;	/* 2 bytes */
	Word		frag;		/* 2 bytes */
	Word		ttlProtocol;	/* 2 bytes */
				/* first byte is TTL = 'time to lose' */
				/* second byte is protocol */
	Word		checksum;	/* 2 bytes */
	IP_Address	source;		/* IP address - 4 bytes */
	IP_Address	destination;	/* IP address - 4 bytes */
};				/* total 20 bytes */

	/* macros which parse parts of the ip message */

#define IP_VERSION(ip)          ( ip->vht >> 12) & 0xf
#define IP_HLEN(ip)             ( ip->vht >> 8) & 0xf
#define IP_HBYTES(ip)           ( ip->vht >> 6) & 0x3c
#define IP_PROTOCOL(ip)  ip->ttlProtocol & 0xff


struct tcp_Header {
	Word		srcPort;
	Word		dstPort;
	Longword	seqnum;
	Longword	acknum;
	Word		flags;
	Word		window;
	Word		checksum;
	Word		urgentPointer;
};

#define TCPF_FIN	0x0001
#define TCPF_SYN	0x0002
#define TCPF_RST	0x0004
#define TCPF_PUSH	0x0008
#define TCPF_ACK	0x0010
#define TCPF_URG	0x0020
#define TCPF_DO		0xF000
#define TCP_DATAOFFSET(tp)  tp->flags >> 12

/* UDP header is only 8 bytes */

/* The TCP/UDP Pseudo Header. Used for computing checksum. */

struct tcp_Pseudoheader {
	IP_Address	src;		/* 4 bytes */
	IP_Address	dst;		/* 4 bytes */
	Byte		mbz;		/* must be zero. 1 byte */
	Byte		protocol;	/* 6 = tcp. 1 byte */
	Word		length;		/* 2 bytes */
	Word		checksum;	/* 2 bytes */
};					/* 14 bytes */


#define TS_LISTEN	1	/* listening for connection */
#define TS_SSYN		2	/* syn sent, active open */
#define TS_RSYN		3	/* syn received, synack+syn sent. */
#define TS_ESTAB	4	/* established */
#define TS_SFIN		5	/* sent FIN */
#define TS_AFIN		6	/* sent FIN, received FINACK */
/* #define TS_CLOSEWT	7 */	/* received FIN waiting for close */
#define TS_RFIN		8	/* (closing) sent FIN, received FIN */
				/* (waiting for FINACK) */
#define TS_LASTACK	9	/* fin received, finack+fin sent */
#define TS_TIMEWT	10	/* dally after sending final FINACK */
#define TS_CLOSED	11	/* (closed) finack received */


/* ------ TCP Socket definition ------------------------------------- */

#define TCP_MAXDATA	512		/* maximum bytes to buffer on output */

struct tcp_Socket {
	struct tcp_Socket * next;	/* pointer to next socket */
	short		state;		/* connection state */
	Procref		dataHandler;	/* called with incoming data */
	struct Ethernet_Address hisethaddr;	/* ethernet address of peer */
	IP_Address	hisaddr;	/* internet address of peer */
	Word		myport, hisport; /* tcp ports for this connection */
	Longword	acknum, seqnum; /* data ack'd and sequence num */
	Longword	timeout;	/* timeout, in milliseconds */
	short		unhappy;	/* flag, retransmitting segt's */
	Word		flags;		/* flags Word for last packet sent */
	short		dataSize;	/* number of bytes of data to send */
	Byte		data[ TCP_MAXDATA ]; /* data to send */
};


int sed_Init P(( Void ));
int sed_Deinit P(( Void ));
Byte * sed_FormatPacket P(( Byte *destEAddr, int ethType ));
int sed_Send P(( int pkLengthInOctets ));
int sed_Receive P(( Byte *recBufLocation ));
Byte * sed_IsPacket P(( Void ));
int sed_CheckPacket P(( Word *recBufLocation, Word expectedType ));
Void tcp_Init P(( Void ));
Void tcp_Open P(( struct tcp_Socket *s, Word lport, IP_Address ina,
	Word port, Procref datahandler ));
Void tcp_Listen P(( struct tcp_Socket *s, Word port, Procref datahandler,
	Longword timeout ));
Void tcp_Close P(( struct tcp_Socket *s ));
Void tcp_Abort P(( struct tcp_Socket *s ));
int tcp P(( Procrefv application ));
int tcp_Write P(( struct tcp_Socket *s, Byte *dp, int len ));
Void tcp_Flush P(( struct tcp_Socket *s ));

	/* redefine move as memcpy */
#define	Move(s,d,n)	memcpy(d,s,n)   
	/* the following are internal to TCP, but because it was broken
		into chunks, they had to become global */

Void tcp_Unthread P(( struct tcp_Socket *ds ));
Void tcp_Retransmitter P(( Void ));
Void tcp_ProcessData P(( struct tcp_Socket *s,
		struct tcp_Header *tp, int len ));
Void tcp_Send P(( struct tcp_Socket *s ));
Void tcp_DumpHeader P(( struct in_Header *ip,
		struct tcp_Header *tp, char *mesg ));
Void tcp_Handler P(( struct in_Header *ip ));

Word checksum P(( Word *dp, int length ));
Longword lchecksum P(( Word *dp, int length ));

Void main P(( int argc, char **argv ));

Longword MsecClock P(( Void ));
void outp_char(char a);
void LineInput(void);
char inp_char();
void counters(void);
void puts (char *s);
void four_bytes(char *buf,unsigned char *n);


#define tcp_RETRANSMITTIME 1000 /* interval at which retransmitter is called */
                                        /* 1 SECOND */
#define tcp_LONGTIMEOUT 30000   /* timeout for opens */
#define tcp_TIMEOUT     20000   /* timeout during a connection */

/* ----- static data ------------------------------------------------ */

static IP_Address              local_IP_address=0,       /* local IP address */
                               remote_IP_address=0;
static int                     tcp_id=0;         /* TCP ID, gets incremented */

static struct tcp_Socket *     tcp_allsocs=0;    /* make global for sharing */
static struct tcp_Socket *     ftp_ctl=0; 
static struct tcp_Socket *     ftp_data=0; 
static Byte * ftp_cmdbuf=0;
static int ftp_cmdbufi=0;

static Byte * ftp_outbuf=0;
static int ftp_outbufix=0, 
           ftp_outbuflen=0;
    
static short ftp_rcvState=0;
static Byte ftp_echoMode=0;
static  unsigned char * p_begin_packet = NULL;
static  unsigned char * p_recv_next = NULL;
static  unsigned char   *xmitbuffer=NULL;
static  unsigned char   *recvbuffer=NULL;
static void * queue=NULL;
static int EndProgram=0;
static char kbrdcycle=0;
static Word remoteport=0;
static HANDLE hreturnfile=0;

#define ftp_StateGETCMD     0       /* get a command from the user */
#define ftp_StateIDLE       1       /* idle connection */
#define ftp_StateINCOMMAND  2       /* command sent, awaiting response */
#define ftp_StateRCVRESP    3       /* received response, need more data */
#define SIZECMDBUF 120
#define OUTBUFSIZE 80


void outp_char(char a)
{
 transmit(&a,1);
}

/* ----- internal definitions --------------------------------------- */

#define	FR_END		0xC0	/* 0300 */
#define	FR_ESC		0xDB	/* 0333 */
#define	T_FR_END	0xDC	/* 0334 */
#define	T_FR_ESC	0xDD	/* 0335 */

#define	BUFFERSIZE	8192


/* ----- send a packet on the link ---------------------------------- */

int sed_Send( pkLengthInBytes ) int pkLengthInBytes; {
	int		i;
	unsigned char *	p_uc;

	if( pkLengthInBytes <= 0 ) return 1;	/* nothing to do */ 

	/* While sending, we have to process the escapes.
		If we find an FR_END, change it to FR_ESC + T_FR_END.
		If we find an FR_ESC, change it to FR_ESC + T_FR_ESC. */

	outp_char( FR_END );		/* send FR_END. */
	p_uc = &xmitbuffer[ 0 ];

	for( i = 0; i < pkLengthInBytes; ++i ) {
		switch( *p_uc ) {
		case FR_END:
			outp_char( FR_ESC );
			outp_char( T_FR_END );
			break;
		case FR_ESC:
			outp_char( FR_ESC );
			outp_char( T_FR_ESC );
			break;
		default:
			outp_char( *p_uc );
		}
		++p_uc;
	}
	outp_char( FR_END );        
	return 1;
}



/* ----- check for packet received ---------------------------- */

	/* Test for the arrival of a packet. The location of the packet is
		returned. If no packet is returned, the routine returns
		zero. (Timeout is irrelevant here.) */

Byte * sed_IsPacket() {
	unsigned char		uc;

#ifdef DEBUG_SED
	unsigned char *		p;
#endif
        static  int     end_count = 0;
        static  int     b_esc_flag = 0;
        static  unsigned char lastuc=0;
        Longword delay;
        if (p_begin_packet==NULL) {
          p_begin_packet = &recvbuffer[0];
          end_count=0;
        }
        if (p_recv_next==NULL) {
          p_recv_next = &recvbuffer[0];
          end_count=0;
        }
        if ((Longword)p_recv_next>
            (Longword)&recvbuffer[BUFFERSIZE-1]) {
               p_begin_packet=&recvbuffer[0];
               p_recv_next=&recvbuffer[0];
               puts("Full buffer");
               OSLinkReset();

              }
              
	/* See if there are more packets in the receiver */
        while( receive(&uc,1) != 0 ) {    /* any chars to receive? RS*/
		if( b_esc_flag ) {
			switch( uc ) {
			case T_FR_ESC:
				*p_recv_next++ = FR_ESC;
				break;
			case T_FR_END:
				*p_recv_next++ = FR_END;
				break;
			default:
				*p_recv_next++ = uc;
			}
			b_esc_flag = 0;
		} else {
			switch( uc ) {
			case FR_ESC:
				b_esc_flag = 1;
				break;
			case FR_END:
				++end_count;
                                if ((end_count==2)&&
                                    (p_begin_packet==p_recv_next)) {
                                        end_count=1;
                                        OSLinkReset();

                                 }

                                if( (end_count == 1)) {
					p_begin_packet = p_recv_next;
                                        if (((Longword)p_begin_packet)&1){
                                               *p_begin_packet++;
                                               }
                                               /*Izbjeci neparne adrese*/
                                }
				break;
			default:
				*p_recv_next++ = uc;
			}
		}
		if( end_count == 2 ) break;
	}
      /*  b_esc_flag=0;*/
	/* do we have a packet in the buffer? */

	if( end_count == 2 ) {
		p_recv_next = &recvbuffer[ 0 ];
		end_count = 0;
		return p_begin_packet;

        } else  {
        /*   end_count=0;  */
           return ( Byte * ) 0;
          }
}


int ResetBuffer()
{
 p_begin_packet=NULL;
 p_recv_next==NULL;

}

/* ----- Format an ethernet header in the transmit buffer ----------- */
	/* return pointer to where to build the IP message */

Byte * sed_FormatPacket( destEAddr, ethType )
	Byte *destEAddr; int ethType; {

	return ( Byte * ) &xmitbuffer[ 0 ];	/* return ptr to 1st Byte */
}

/* ----- busy-wait loop for tcp. Also calls an "application proc" --- */

int tcp( application )

	Procrefv application;
{
             struct in_Header *      ip;
             Longword                timeout, start;


	timeout = 0L;
	for(;;) {
               if (EndProgram) break;
		start = ( Longword ) MsecClock();
		/* Have we received a packet? */
		ip = ( struct in_Header * ) sed_IsPacket();
		if( ip == ( struct in_Header * ) 0 ) {
			/* No, we haven't. */
			if( MsecClock() > timeout ) {
				/* Anything to retransmit? */
				tcp_Retransmitter();
				/* Set next transmit time */ 
        			timeout = MsecClock()	+ tcp_RETRANSMITTIME;
			}
                      /* There's nothing to do. Let the user enter a command. */
			application();
			continue;
		}

		/* We have received a packet. Process it */
               if(  ip -> destination  != local_IP_address ) {
#ifdef DEBUG_TCP
                                puts( "Wrong IP" );
#endif
                                OSLinkReset();
                                ResetBuffer();
                                goto recycle;
		}
               if( checksum( ( Word * ) ip, IP_HBYTES( ip )) != 0xFFFF ) {
#ifdef DEBUG_TCP
                                puts("Bad IP checksum!" );
#endif
                                OSLinkReset();
                                ResetBuffer();
                                goto recycle;
		}

                   if ( (IP_PROTOCOL( ip )) == 6) tcp_Handler(ip);

		/* recycle buffer */
recycle:
	}

	return 1;	
}


/* ----- compute a Longword sum of words in message ----------------- */
	/* (partial checksum) */

Longword lchecksum( dp, length ) Word *dp; int length; {
             int             len;
             Longword        sum;

	len = length >> 1;
	sum = 0L;

        while ( len-- > 0 ) sum += ( *dp++ );
        if( length & 1 ) sum += ( ( *dp ) & 0xFF00 );

	return sum;
}


/* ----- calculate Word checksum ------------------------------------ */

Word checksum( dp, length ) Word *dp; int length; {
             Longword        sum;

	sum = lchecksum( dp, length );

	while( sum & 0xFFFF0000L )
		sum = ( sum & 0xFFFFL ) + (( sum >> 16L ) & 0xFFFFL );

	return ( Word )( sum & 0xFFFF );
}



Void tcp_Send( s ) struct tcp_Socket *s; {
             struct tcp_Pseudoheader ph;
             struct _pkt {
			struct in_Header	in;
			struct tcp_Header	tcp;
			Longword		maxsegopt;
		} * pkt;
             Byte *                  dp;
             Longword                lw;

	/* don't do it if the state is Closed or the socket is noton the linklist */

	if(( s -> state == 0 ) || ( s -> state == TS_CLOSED ))
		return;
	pkt = ( struct _pkt * ) sed_FormatPacket(
		( Byte * ) &( s -> hisethaddr.w[ 0 ] ), 0x800 );
	dp = ( Byte * ) &( pkt -> maxsegopt );
	if( s -> flags & TCPF_SYN ) {
		/* Should not send data on SYN */
                pkt -> in.length = ( sizeof( struct in_Header )
			+ sizeof( struct tcp_Header ));
	} else {
                pkt -> in.length = ( sizeof( struct in_Header )
			+ sizeof( struct tcp_Header ) + s -> dataSize );
	}
	/* tcp header */
        pkt -> tcp.srcPort = ( s -> myport );
        pkt -> tcp.dstPort = ( s -> hisport );
        pkt -> tcp.seqnum = ( s -> seqnum );
        pkt -> tcp.acknum = ( s -> acknum );
        pkt -> tcp.window = ( 64 );
		/* This 64 is the size of data which I am willing to receive */
        pkt -> tcp.flags = ( s -> flags | 0x5000 );
	pkt -> tcp.checksum = 0;
	pkt -> tcp.urgentPointer = 0;
	if( s -> flags & TCPF_SYN ) {
		/* If sending the options, add 1 DWORD to the data offset */
                pkt -> tcp.flags = ( ( pkt -> tcp.flags ) + 0x1000 );
		/* Add 4 to the length */
                pkt -> in.length = ( ( pkt -> in.length ) + 4 );
		/* Options. This is really: kind 02, length 04, value 1400.*/
                pkt -> maxsegopt = ( 0x02040578 ); /* 1400 bytes */
		dp += 4;
	} else {
		/* Only send the data if NOT a SYN. */
                Move( s -> data, dp, s -> dataSize );
	}

	/* internet header */
        pkt -> in.vht = ( 0x4500 );	/* version 4, hdrlen 5, tos 0 */
        pkt -> in.identification = ( tcp_id++ );
	pkt -> in.frag = 0;
        pkt -> in.ttlProtocol = (( 250 << 8 ) + 6 );
	pkt -> in.checksum = 0;
        pkt -> in.source = ( local_IP_address );
        pkt -> in.destination = ( s -> hisaddr );
        pkt -> in.checksum = ((Word ) ~ checksum( ( Word * ) &pkt -> in,
				sizeof( struct in_Header )));
	/* compute tcp checksum */
	ph.src = pkt -> in.source;
	ph.dst = pkt -> in.destination;
	ph.mbz = 0;
	ph.protocol = 6;
        ph.length = (( pkt -> in.length ) - sizeof( struct in_Header ));
	lw = lchecksum( ( Word * ) &ph, sizeof( ph ) - 2 );
	while( lw & 0xFFFF0000L )
		lw = ( lw & 0xFFFFL ) + (( lw >> 16L ) & 0xFFFFL );
	lw += lchecksum( ( Word * ) &pkt -> tcp,
                ( ph.length ));
	while( lw & 0xFFFF0000L )
		lw = ( lw & 0xFFFFL ) + (( lw >> 16L ) & 0xFFFFL );
        pkt -> tcp.checksum = ( ~ ( Word )( lw & 0xFFFFL ));
#ifdef DEBUG_TCP
	tcp_DumpHeader( &pkt -> in, &pkt -> tcp, "Sending" );
#endif
        sed_Send( ( pkt -> in.length ));
}


/* ----- retransmitter ---------------------------------------------- */

/* Retransmitter - called periodically to perform tcp retransmissions */

Void tcp_Retransmitter( Void ) {
             struct tcp_Socket *     s;
             short                   x;
	/* Go through all of the open sockets and see if there is
		anything to send */
	for( s = tcp_allsocs; s; s = s -> next ) {
		x = False;
		if( s -> dataSize > 0 || s -> unhappy ) {
			/* send the data */
			tcp_Send( s );
			x = True;
		}
		/* anything sent? */
		if( x || s -> state != TS_ESTAB )
			s -> timeout -= tcp_RETRANSMITTIME;
		if( s -> timeout <= 0 ) {
			if( s -> state == TS_TIMEWT ) {
#ifdef DEBUG_TCP
                                puts("Closed." );
#endif
                                s -> state = TS_CLOSED;
				if( s -> dataHandler != 0 )
					( s -> dataHandler )(( void * ) s,
						( Byte * ) 0, 0 );
#ifdef DEBUG_TCP
                                else  puts("got close no handler" );
#endif

				/* Unthread this socket from the linklist */
				tcp_Unthread( s );
			} else {
                                puts( "Timeout\n" );
				tcp_Abort( s );
			}
		}
	}
}


Longword MsecClock P(( Void ))
{ Longword value;
  if (OSTimerExpired(6)) OSTimerRestart(6);
  value=5*(32000-OSTimerCurVal(6));
  return (value);
}


/* ----- Unthread a socket from the socket list, if it's there ------ */

Void tcp_Unthread( ds ) struct tcp_Socket *ds; {
             struct tcp_Socket *s, **sp;
	if( ds == ( struct tcp_Socket * ) 0 ) return;
	sp = &tcp_allsocs;			/* -> -> socket */
	for(;;) {
		s = *sp;			/* what pointer points to */
		if( s == ( struct tcp_Socket * ) 0 ) break;	/* end of list? */
		if( s == ds ) {			/* matches one to delete? */
			*sp = s -> next;	/* unlink it from the list */
			break;			/* all done */
		}
		sp = &s -> next;		/* move to the next one */
	}
	ds -> next = ( struct tcp_Socket * ) 0;	/* clear next pointer */
}


/* ----- Write data to a connection. -------------------------------- */
 
	/* Returns number of bytes written, == 0 when connection is not in
 		established state. */

int tcp_Write( s, dp, len )
	struct tcp_Socket *s;
	Byte *dp;
	int len;
{
             int     x;

	/* if the connection is not established, don't send anything */

	if( s -> state != TS_ESTAB ) return 0;

	/* figure out how much I can move */

	if( len > ( x = TCP_MAXDATA - s -> dataSize ) ) len = x;
	if( len > 0 ) {
                Move( dp, &s -> data[ s -> dataSize ], len );
		s -> dataSize += len;
		tcp_Flush( s );
	}
	return len;	/* return count of bytes sent */
}

/* ----- Send pending data ------------------------------------------ */
Void tcp_Flush( s ) struct tcp_Socket *s; {
	if(( s -> state == 0 ) || ( s -> state == TS_CLOSED ))
		return;
	if( s -> dataSize > 0 ) {
		s -> flags |= TCPF_PUSH;
		tcp_Send( s );
	}
}



/* ----- Dump tcp protocol header of a packet ----------------------- */
#ifdef DEBUG_TCP
Void tcp_DumpHeader( ip, tp, mesg )
	struct in_Header *ip;
	struct tcp_Header *tp;
	char *mesg;
{
        static char *flags[] = { "FIN", "SYN", "RST", "PUSH", "ACK", "URG" };
             int             len;
             Word            f;

        len = ( ip -> length )
		- (( TCP_DATAOFFSET( tp ) + IP_HLEN( ip )) << 2 );

        printf("TCP: %s pkt", mesg);
        printf("SP: %x; DP: %x; ",
                ( tp -> srcPort ),
                ( tp -> dstPort ));
        printf("Seq=%lx AckN%lx Wnd=%d DLen=%d",
                ( tp -> seqnum ),
                ( tp -> acknum ),
                ( tp -> window ),
		len );
        printf("DO=%d, C=%x U=%d",
		TCP_DATAOFFSET( tp ),
                ( tp -> checksum ),
                ( tp -> urgentPointer ));


        f = ( tp -> flags );
        printf("%x",f);
  
}
#endif


/* ----- Handler for incoming packets. ------------------------------ */

Void tcp_Handler( ip ) struct in_Header *ip; {
             struct tcp_Header *     tp;
             struct tcp_Pseudoheader ph;
             int                     len;
             int                     diff;
             struct tcp_Socket *     s;
             Word                    flags;
             Longword                lw;
	len = IP_HBYTES( ip );
	tp = ( struct tcp_Header * )(( Byte * ) ip + len );
        len = ( ip -> length ) - len;
	/* demux to active sockets */
	for( s = tcp_allsocs; s; s = s -> next )
		if(( s -> hisport != 0 )
                        && ( ( tp -> dstPort ) == s -> myport )
                        && ( ( tp -> srcPort ) == s -> hisport )
                        && ( ( ip -> source )
				== s -> hisaddr ))
			break;
	if( s == 0 ) {	/* didn't find an matching socket */
		/* demux to passive sockets */
		for ( s = tcp_allsocs; s; s = s -> next )
			if(( s -> hisport == 0 )
                                && ( ( tp -> dstPort )
					== s -> myport ))
				break;
	}
	if( s == 0 ) {	/* Still didn't find a matching socket */
#ifdef DEBUG_TCP
		tcp_DumpHeader( ip, tp, "Discarding" );
#endif
		return;
	}
#ifdef DEBUG_TCP
	tcp_DumpHeader( ip, tp, "Received" );
#endif
	ph.src = ip -> source;
	ph.dst = ip -> destination;
	ph.mbz = 0;
	ph.protocol = 6;
        ph.length = ( len );
	lw = lchecksum( ( Word * ) &ph, sizeof ph - 2 );
	while( lw & 0xFFFF0000L )
		lw = ( lw & 0xFFFFL ) + (( lw >> 16L ) & 0xFFFFL );
	lw += lchecksum( ( Word * ) tp, len );
	while( lw & 0xFFFF0000L )
		lw = ( lw & 0xFFFFL ) + (( lw >> 16L ) & 0xFFFFL );
        if( lw != 0x0000FFFFL ) {
#ifdef DEBUG_TCP
                printf( "bad tcp checksum (%lx)", lw );
#endif
                OSLinkReset();
                ResetBuffer(); 
                return; 
                }               
        flags = ( tp -> flags );
	if( flags & TCPF_RST ) {
#ifdef DEBUG_TCP
                puts("reset\n");
#endif
		s -> state = TS_CLOSED;
		if( s -> dataHandler != 0 )		/* cast removed */
			( s -> dataHandler )(( void * ) s, ( Byte * ) 0, -1);
#ifdef DEBUG_TCP
                else puts( "got close" );
#endif
		tcp_Unthread( s );
		return;
	}
	switch( s -> state ) {

	case TS_LISTEN:
		/* Initial state of a Listen port */
	if( flags & TCPF_SYN ) {
			/* We expect to get Syn back */
       	               s -> acknum = ( tp -> seqnum ) + 1;
       	               s -> hisport = ( tp -> srcPort );
       	               s -> hisaddr = ( ip -> source );
		s -> flags = TCPF_SYN | TCPF_ACK;
			tcp_Send( s );
			s -> state = TS_RSYN;
			s -> unhappy = True;
			s -> timeout = tcp_TIMEOUT;
#ifdef DEBUG_TCP
                        puts("Syn");
#endif
		}
		break;

	case TS_SSYN:
		/* Initial state of a Open port */
		if( flags & TCPF_SYN ) {
			s -> acknum++;
			s -> flags = TCPF_ACK;
			s -> timeout = tcp_TIMEOUT;
		if(( flags & TCPF_ACK )
                                && ( tp -> acknum )
					== ( s -> seqnum + 1 ) ) {
             		 	puts( "--- Open! ---\n" );
  			s -> state = TS_ESTAB;
				s -> seqnum++;
                              s -> acknum = ( tp -> seqnum )+ 1;
				s -> unhappy = False;
			} else {
				s -> state = TS_RSYN;
			}
		} else {
#ifdef DEBUG_TCP
                        puts( "Sent Syn, didn't get Syn back\n" );
#endif
			/* This sounds to rr like the half-open
				condition in fig. 10, p. 74 (sec. 3.4). */
			s -> flags = TCPF_RST;
       	               s -> seqnum = ( tp -> acknum );
		tcp_Send( s );
#ifdef DEBUG_TCP
               puts("Sent RST!\n" );
#endif
			/* We expect no response. Reset the flags to send
				SYN when the timeout occurs. */
			s -> flags = TCPF_SYN;
			s -> timeout = tcp_TIMEOUT;
			s -> seqnum = 0;	/* Start over */ 
			/* Stay in SYNSENT and wait for timeout. */
		}
		break;

	case TS_RSYN:
		if( flags & TCPF_SYN ) {
			s -> flags = TCPF_SYN | TCPF_ACK;
			tcp_Send(s);
			s -> timeout = tcp_TIMEOUT;
#ifdef DEBUG_TCP
                        puts(" retransmit of original syn\n");
#endif
		}
		if(( flags & TCPF_ACK )
                        && ( tp -> acknum )
				== ( s -> seqnum + 1L ) ) {
			s -> flags = TCPF_ACK;
			tcp_Send( s );
			s -> seqnum++;
			s -> unhappy = False;
			s -> state = TS_ESTAB;
			s -> timeout = tcp_TIMEOUT;
#ifdef DEBUG_TCP
                        puts( "Synack received - connection established\n" );
#endif

		} else {
#ifdef DEBUG_TCP
                        puts( "Wrong syn. ");
#endif
		}
		break;

	case TS_ESTAB:
		if(( flags & TCPF_ACK ) == 0 ) return;
		/* process ack value in packet */
                diff = ( int )( ( tp -> acknum )- s -> seqnum );
		if( diff > 0 ) {
			/* The diff value is the number of bytes of MY data
				which he is acknowledging. */
			if( diff > TCP_MAXDATA ) diff = TCP_MAXDATA;
			else {
                                Move( &s -> data[ diff ],
					&s -> data[ 0 ],
                                        TCP_MAXDATA - diff );
			}

			s -> dataSize -= diff;	/* bytes to send remaining */
			s -> seqnum += diff;	/* my next sequence number */
		}
		s -> flags = ( Word ) TCPF_ACK;
		tcp_ProcessData( s, tp, len );
		break;

	case TS_SFIN:
		if(( flags & TCPF_ACK ) == 0 ) return;
                diff = ( int )( ( tp -> acknum )
			- s -> seqnum - 1 );
		s -> flags = ( Word )( TCPF_ACK | TCPF_FIN );
		if( diff == 0 ) {
			s -> state = TS_AFIN;
			s -> flags = TCPF_ACK;
#ifdef DEBUG_TCP
                        puts("finack received.\n");
#endif
		}
		tcp_ProcessData(s, tp, len);
		break;

	case TS_AFIN:
		s -> flags = TCPF_ACK;
		tcp_ProcessData( s, tp, len );
		break;

	case TS_RFIN:
                if( ( tp -> acknum ) == (s -> seqnum + 1) ) {
			s -> state = TS_TIMEWT;
			s -> timeout = tcp_TIMEOUT;
		}
		break;

	case TS_LASTACK:
                if( ( tp -> acknum ) == (s -> seqnum + 1) ) {
			s -> state = TS_CLOSED;
			s -> unhappy = False;
			s -> dataSize = 0;
			if( s -> dataHandler != 0 )	/* cast removed */
				( s -> dataHandler )(( void * ) s,
					( Byte * ) 0, 0 );
#ifdef DEBUG_TCP
                      else    puts( "got close, no handler\n" );
#endif
			tcp_Unthread( s );
#ifdef DEBUG_TCP
                      puts( "Closed. [626]\n" );
#endif
		} else {
			s -> flags = TCPF_ACK | TCPF_FIN;
			tcp_Send( s );
			s -> timeout = tcp_TIMEOUT;
#ifdef DEBUG_TCP
                        puts("retransmitting FIN\n" );
#endif
		}
		break;

	case TS_TIMEWT:
		s -> flags = TCPF_ACK;
		tcp_Send(s);
	}
}


/* ----- Process the data in an incoming packet. -------------------- */

/* Called from all states where incoming data can be received: established,
	fin-wait-1, fin-wait-2 */

Void tcp_ProcessData( s, tp, len )
	struct tcp_Socket *s;
	struct tcp_Header *tp;
	int len;
{
             int             diff, x;
             Word            flags;
             Byte *          dp;
        flags = ( Word ) ( tp -> flags );
	/* Look at the difference between the value I'm acking and the
		sequence number he's sending. The difference (if any) is the
		length of data in the buffer which we have already seen. */
        diff = ( int )( s -> acknum - ( tp -> seqnum ));
	/* I don't understand the following statement, unless it is supposed
		to be compensating for options. There should not be data
		on a SYN packet. */
	if( flags & TCPF_SYN ) diff--;
	x = TCP_DATAOFFSET( tp ) << 2;	/* mult. times 4 */
	dp = (( Byte * ) tp ) + x;		/* point to data */
len -= x;				/* subtract offset */
	if( diff >= 0 ) {
		dp += diff;		/* Ignore stuff we've already seen */
		len -= diff;		/* subtract length of data we've already seen */
		s -> acknum += len;	/* Add len to acknum to acknowledge new data */
		if( s -> dataHandler != 0 )	/* cast removed */
			( s -> dataHandler )(( void * ) s, dp, len );
#ifdef DEBUG_TCP
                else    printf( "got data, %d bytes\n", len );
#endif
		if( flags & TCPF_FIN ) {
			s -> acknum++;
#ifdef DEBUG_TCP
                        puts( "consumed fin.\n" );
#endif
			switch(s -> state) {
			case TS_ESTAB:
			/* note: skip state CLOSEWT by automatically closing conn */
				x = TS_LASTACK;
				s -> flags |= TCPF_FIN;
				s -> unhappy = True;
#ifdef DEBUG_TCP
                                puts( "sending fin.\n" );
#endif
				break;
			case TS_SFIN:
				x = TS_RFIN;
				break;
			case TS_AFIN:
				x = TS_TIMEWT;
				break;
			}
			s -> state = x;
		}
	} else {
#ifdef DEBUG_TCP
                printf( "diff was negative, %d\n", diff );
#endif
	}
	s -> timeout = tcp_TIMEOUT;
	tcp_Send( s );
}


/* ----- tcp open --------------------------------------------------- */

/* Actively open a TCP connection to a particular destination. */

Void tcp_Open( s, lport, ina, port, datahandler )
	struct tcp_Socket *	s;
	IP_Address		ina;
	Word			lport, port;

	Procref			datahandler;
{


	s -> state = TS_SSYN;
	s -> timeout = tcp_LONGTIMEOUT;
	if( lport == 0 ) lport = ( Word ) MsecClock();
	s -> myport = lport;
	s -> hisaddr = ina;
	s -> hisport = port;
	s -> seqnum = 0;
	s -> dataSize = 0;
	s -> flags = TCPF_SYN;		/* Looking for sync */
	s -> unhappy = True;		/* Flag it as unhappy */
	s -> dataHandler = datahandler;
	/* add the new socket to the linklist */
	s -> next = tcp_allsocs;
	tcp_allsocs = s;
	tcp_Send( s );
}


/* ----- tcp listen ------------------------------------------------- */

/* Passive open: listen for a connection on a particular port */

Void tcp_Listen( s, port, datahandler, timeout )
	struct	tcp_Socket *	s;
	Word			port;

	Procref			datahandler;
	Longword		timeout;
{
	s -> state = TS_LISTEN;

	if( timeout == 0L ) s -> timeout = 0x7FFFFFFL; /* forever... */
	else	s -> timeout = timeout;

	s -> myport = port;
	s -> hisport = 0;
	s -> seqnum = 0;
	s -> dataSize = 0;
	s -> flags = 0;
	s -> unhappy = 0;
	s -> dataHandler = datahandler;

	/* Add it to the linklist */

	s -> next = tcp_allsocs;
	tcp_allsocs = s;
}


/* ----- tcp close -------------------------------------------------- */

/* Send a FIN on a particular port -- only works if it is open */

Void tcp_Close( s ) struct tcp_Socket *s; {
	if( s -> state == TS_ESTAB || s -> state == TS_RSYN ) {
		s -> flags = TCPF_ACK | TCPF_FIN;
		s -> state = TS_SFIN;
		s -> unhappy = True;
	}
}

/* ----- Abort a tcp connection ------------------------------------- */

Void tcp_Abort( s ) struct tcp_Socket *s; {

	if( s -> state != TS_LISTEN && s -> state != TS_CLOSED ) {
		s -> flags = TCPF_RST | TCPF_ACK;
		tcp_Send( s );
	}

	s -> unhappy = 0;
	s -> dataSize = 0;
	s -> state = TS_CLOSED;

	if( s -> dataHandler != 0 )		/* cast to Procref removed */
		( s -> dataHandler )(( void * ) s, ( Byte * ) 0, -1 );
#ifdef DEBUG_TCP
        else    puts( "got abort, no handler\n" );
#endif
	tcp_Unthread( s );
}



static int __x = 0, __y = 0;

void putc (char c)
{
  int h = 6 + 2 * FontGetSys ();
  if (c == '\n' || __x + FontCharWidth (c) > ScrRect->xy.x1) __x = 0, __y += h;
  if ( __y + h > ScrRect->xy.y1) ScrRectScroll (ScrRect, ScrRect, h, A_REVERSE), __y -= h;
  if (c==127) {
    __x -= FontCharWidth (c);
    if (__x<0) { __y -=h;
        __x=ScrRect->xy.x1;
        if (__y<0) __y=0;
    }
  } else
  if (c != '\n') DrawChar (__x, __y, c, A_REPLACE), __x += FontCharWidth (c);
}

void puts (char *s)
{
  int i;
  for (i = 0; i < strlen(s); i++) putc (s[i]);
}
#define printf(format,args...) ({char buffer[200]; sprintf (buffer, format, ##args); puts (buffer);})
#define clrscr() (__x = __y = 0, ClrScr ())



HANDLE CreateFile(char *FileName) // Vraca handle, H_NULL u slucaju greske
{
  HANDLE h;
  SYM_ENTRY *sym_entry;
  char str[30],*sptr=str;
  *sptr=0; while(*++sptr=*FileName++);
  if(!(h=HeapAlloc(HeapMax()))) return H_NULL;
  if(!(sym_entry=DerefSym(SymAdd(sptr)))) return H_NULL;
  *(long*)HLock(sym_entry->handle=h)=0x00010000;
  printf ("Handle %d\n",h);
  return h;
}

void AppendCharToFile(HANDLE h,unsigned char c)
{
  char *base=HeapDeref(h);
  unsigned len=*(unsigned*)base;
  if(len>HeapSize(h)-10) return;
  *(unsigned*)base=len+1;
  base[len+2]=c;
}

void AppendBlockToFile(HANDLE h,void *addr,unsigned int len)
{
  unsigned int i;
  for(i=len;i;i--) AppendCharToFile(h,*((char*)addr)++);
}

void CloseFile(HANDLE h)
{
  printf ("Close Handle %d\n",h);
  AppendCharToFile(h,0); AppendCharToFile(h,0x2D);
  HeapUnlock(h);
  HeapRealloc(h,*(unsigned*)HeapDeref(h)+3);
}


ftp_Abort()
{
    tcp_Abort(ftp_ctl);
    tcp_Abort(ftp_data);
}



void ftp_ctlHandler(s, dp, len)
    struct tcp_Socket *s;
    Byte *dp;
    int len;
{
    Byte c, *bp, data[80];
    int i;

    if ( dp == 0 ) {
        tcp_Abort(ftp_data);
        return;
    }

    do {
        i = len;
        if ( i > sizeof data ) i = sizeof data;
        Move(dp, data, i);
        len -= i;
        bp = data;
        while ( i-- > 0 ) {
            c = *bp++;
            if ( c != '\r' ) {
                if ( c == '\n' ) {
                    ftp_cmdbuf[ftp_cmdbufi] = 0;
                    if (remoteport==21)
                      ftp_commandLine();
                    else 
                      printf("> %s\n", ftp_cmdbuf);  
                    ftp_cmdbufi = 0;
                } else if ( ftp_cmdbufi < (SIZECMDBUF)-1 ) {
                    ftp_cmdbuf[ftp_cmdbufi++] = c;
                }
            }
        }
    } while ( len > 0 );
}


ftp_commandLine()
{
    printf("> %s\n", ftp_cmdbuf);
    switch(ftp_rcvState) {
     case ftp_StateIDLE:
        if ( ftp_cmdbuf[3] == '-' )
            ftp_rcvState = ftp_StateRCVRESP;
        break;

     case ftp_StateINCOMMAND:
        if ( ftp_cmdbuf[3] == '-' )
            ftp_rcvState = ftp_StateRCVRESP;
     case ftp_StateRCVRESP:
        if ( ftp_cmdbuf[3] == ' ' )
            ftp_rcvState = ftp_StateIDLE;
        break;
    }
}

void ftp_application(void)
{
    char *s;
    char *dp;
    char recv_filename[8];
    int i,j,putsx,putsy;
    i = -1;
    if ((kbrdcycle++ == 0 )||(ftp_rcvState==ftp_StateGETCMD))
      if ( !OSdequeue(&i,queue) ) {
          if ( i == 264 ) {

              printf("Closing...\n");
              if ( hreturnfile)
                  CloseFile(hreturnfile);

              tcp_Close(ftp_ctl);
              EndProgram=1;
          }
      }

    switch (ftp_rcvState) {
      case ftp_StateGETCMD:
 getcmd:
       if ( i != -1 ) {
            ftp_outbuf[ftp_outbuflen] = 0;
            if (i==176) i=64;
            if (i==168) i=35;
            switch (i) {
                case 257:
                    if ( ftp_outbuflen > 0 ) {
                        ftp_outbuflen--;
                        putc(127);
                        putc(' ');
                        
                    }
                    break;
                case '\r':
                case '\n':
                    puts("\n");
                    dp = &ftp_outbuf[ftp_outbuflen];
                    goto docmd;

                default:
                    if ( i >= ' ' && ftp_outbuflen < OUTBUFSIZE ) {
                        ftp_outbuf[ftp_outbuflen++] = i;
                        putc(i);
                    }
            }
        }
        break;
      case ftp_StateIDLE:
            ftp_rcvState = ftp_StateGETCMD;
            ftp_outbuflen = 0;
            puts("Command> ");
            goto getcmd;
 docmd: 
        *dp++ = '\r';
        *dp++ = '\n';
        if(( strncmp( ftp_outbuf, "RETR ", 5 ) == 0 )
       	|| ( strncmp( ftp_outbuf,"retr ", 5 ) == 0 )){
             if (! hreturnfile)
                hreturnfile=CreateFile("received");
        }
        if(( strncmp( ftp_outbuf, "LIST ", 5 ) == 0 )
        || ( strncmp( ftp_outbuf,"list ", 5 ) == 0 )){
                hreturnfile=0;
        }
        ftp_outbuflen = (unsigned long)dp - (unsigned long)ftp_outbuf;
        ftp_outbufix = 0;
        ftp_rcvState = ftp_StateINCOMMAND;
        /* fall through */
    case ftp_StateINCOMMAND:
        i = ftp_outbuflen - ftp_outbufix;
        if ( i > 0 ) {
            i = tcp_Write(ftp_ctl, &ftp_outbuf[ftp_outbufix], i);
            ftp_outbufix += i;
            tcp_Flush(ftp_ctl);
        }
        /* fall through */
    case ftp_StateRCVRESP:
        break;
    }
}


void ftp_dataHandler(s, dp, len)
    struct tcp_Socket *s;
    Byte *dp;
    int len;
{
   if (!(hreturnfile)) 
     while (len>0) {
      if((*dp<32) || (*dp>126))
          printf (" %02x ",*dp);
      else   putc( *dp );
      ++dp;
      --len;
    }
    else {
      AppendBlockToFile(hreturnfile,dp,len);
      dp+=len;
      len=0;
      putc('#');
    }                                                                                        }


void TelnetHandler(s, dp, len)
    struct tcp_Socket *s;
    Byte *dp;
    int len;
{
    Byte c, *bp, data[80];
    int i;


    do {
        i = len;
        if ( i > sizeof data ) i = sizeof data;
        Move(dp, data, i);
        len -= i;
        bp = data;
        while ( i-- > 0 ) {
            c = *bp++;
            if ((c=='\n')) printf("\n");
              else if (c>=' ') putc(c);
        }
    } while ( len > 0 );

}


void telnet_application(void)
{   int i,n;
    i = -1;
    if (kbrdcycle++ == 0 )
      if ( !OSdequeue(&i,queue) ) {
          if ( i == 264 ) {
              printf("Closing...\n");
              tcp_Close(ftp_ctl);
              EndProgram=1;
          }
      }
       if ( i != -1 ) {
            if (i==176) i=64;
            if (i==168) i=35;
            ftp_outbuf[1]=i;
            ftp_outbuflen=1;
            putc(i);
            n = tcp_Write(ftp_ctl, &ftp_outbuf[1],1);
            tcp_Flush(ftp_ctl);
        }
}


applayer(host, dataHandler,protocol)
    IP_Address host;
    Procref dataHandler;
    int protocol;
{
    Word port;
    char filecmd[80];

    port = 0x4200;
    /* set up state variables */
    ftp_rcvState = ftp_StateRCVRESP;
    ftp_cmdbufi = 0;
    switch (protocol) { 
       case 1:
           tcp_Listen(ftp_data, 20, dataHandler, 0);
           remoteport=21;
           break;
       case 2: remoteport=25;break;
       case 3: remoteport=110;break;
       case 4: remoteport=80;break;
       case 5: remoteport=23;
           tcp_Open(ftp_ctl, port, host, remoteport, TelnetHandler);
           break;
     }  
     if (remoteport !=23)
            tcp_Open(ftp_ctl, port, host, remoteport, ftp_ctlHandler);
     if (remoteport !=23) tcp(ftp_application);
        else tcp(telnet_application);
}


unsigned long parse_ip(char *s)
{
  unsigned char c;
  unsigned long b,r=0;
  unsigned int i;
  for(i=0;i<=3;i++)
    {
      c=*s;
      if(!isdigit(c)) return 0;
      b=0;
      while(isdigit(c))
        {
          b=10*b+c-'0';
          c=*++s;
        }
      if(b>255) return 0;
      if(!((c=='.' && i<3)||(c==0 && i==3))) return 0;
      s++;
      r=256*r+b;
    }
  return r;
}



int connect_dialog(unsigned long *local,unsigned long *remote,
  unsigned char *email,unsigned int *protocol)
{
  HANDLE dhandle,chandle;
  char reqbuffer[200];
  int combobuffer[1];
  int i,key;
  static char *opc[5]={"FTP","SMTP","POP3","HTTP","TELNET"};
  static char *edt[3]={"Local:","Remote:","My email:"};
  four_bytes(reqbuffer,(char*)local);
  four_bytes(reqbuffer+70,(char*)remote);
  strcpy(reqbuffer+140,email);
  combobuffer[0]=*protocol;
  dhandle=DialogNew(150,70,NoCallBack);
  DialogAddTitle(dhandle,"CONNECT TO REMOTE HOST",BT_OK,BT_CANCEL);
  for(i=0;i<3;i++) DialogAddRequest(dhandle,5,15+10*i,edt[i],70*i,60,15);
  chandle=PopupNew(NULL,44);
  for(i=0;i<5;i++) PopupAddText(chandle,-1,opc[i],0);
  DialogAddPulldown(dhandle,5,45,"Protocol:",chandle,0);
  key=DialogDo(dhandle,CENTER,CENTER,reqbuffer,combobuffer);
  HeapFree(dhandle); HeapFree(chandle);
  if(key!=13) return 0;
  *local=parse_ip(reqbuffer);
  *remote=parse_ip(reqbuffer+70);
  if(*local==0 || *remote==0) return 0;
  strcpy(email,reqbuffer+140);
  *protocol=combobuffer[0];
  return 1;
}

void four_bytes(char *buf,unsigned char *n)
{
  sprintf(buf,"%d.%d.%d.%d",peek(n),peek(n+1),peek(n+2),peek(n+3));
}


/* ----- Initialize the tcp implementation -------------------------- */

Void tcp_Init( Void ) {
        tcp_allsocs = ( struct tcp_Socket * ) 0;
        tcp_id = 0;
        p_begin_packet = NULL;
        p_recv_next = NULL;

}


_main(){
  unsigned int protocol=3;
  char *email="megaribi@bih.net.ba";

  LCD_BUFFER savescr;
  LCD_save(savescr);
  EndProgram=0;
  queue=kbd_queue();
  clrscr();
  OSLinkReset();
  OSFreeTimer(6);
  OSRegisterTimer(6,32000);
  xmitbuffer=malloc(BUFFERSIZE);
  recvbuffer=malloc(BUFFERSIZE);
  ftp_ctl=malloc(sizeof(struct tcp_Socket));
  ftp_data=malloc(sizeof(struct tcp_Socket));
  ftp_cmdbuf=malloc(SIZECMDBUF);
  ftp_outbuf=malloc(OUTBUFSIZE);
  while (connect_dialog(&local_IP_address, &remote_IP_address,email,&protocol)==1)
     {   EndProgram=0;
         hreturnfile=0;
         FontSetSys(F_4x6);
         tcp_Init();
         applayer(remote_IP_address,ftp_dataHandler,protocol);
     }
  free(xmitbuffer);
  free(recvbuffer);
  free(ftp_ctl);
  free(ftp_data);
  free(ftp_cmdbuf);
  free(ftp_outbuf);

  LCD_restore(savescr);
  OSFreeTimer(6);
}

